//	SHM.C:  Shape manager for

//	CGA/EGA/VGA/CGA GRAY/EGA GRAY

//	Handles the following functions:
//		shm_init :Init shape manager
//		shm_do   :Prepare shape tables (load/trash) using shm_want[]
//		shm_exit :Terminate shape manager

#include <stdlib.h>;
#include <io.h>;
#include <alloc.h>;
#include <fcntl.h>;
#include <string.h>;
#include <math.h>;
#include <mem.h>;
#include "\develop\xargon\include\gr.h";

extern void rexit (int num);

int shm_want[shm_maxtbls];
char *shm_tbladdr[shm_maxtbls];
int shm_tbllen[shm_maxtbls];
int shm_flags[shm_maxtbls];

char shm_fname[80];
char colortab[256];

void shm_init (char *fname) {
	int c;

	strcpy (shm_fname, fname);
	for (c=0; c<shm_maxtbls; c++) {
		shm_want[c]=0;
		shm_tbladdr[c]=NULL;
		};
	};

void init8bit (void) {					// Init pallette for numcolorbits=8
	int x;

	for (x=0; x<256; x++) {
		colortab[x]=x;
		};
	};

// K1n9_Duk3 mod: Group the shape header variables into a struct, so that we
// can read the whole header with a single read function call.
typedef struct {
	unsigned char numshapes;
	int numrots;
	int len_cga;
	int len_ega;
	int len_vga;
	char numcolorbits;
	int flags;
} shapeheadertype;

// K1n9_Duk3 mod: Do the same for the image header:
typedef struct {
	byte xs;
	byte ys;
	byte storetype;
} imageheadertype;

#if sizeof(imageheadertype) != 3
#error Wrong imageheadertype size, turn off word alignment!
#endif

#define readin(dest,addr,len) memcpy (&dest,addr,len); addr+=len
#define wr(addr,ofs,src,len) memcpy (addr+ofs,&src,len); ofs+=len
#define wr1(addr,ofs,byt) *(addr+(ofs++))=byt

// K1n9_Duk3 mod: Pass file handle instead of pointers.
void xlate_table (int handle, int n/*, char *data, char *bucket1*/) {
	/*
	unsigned char numshapes=0;
	char numcolorbits=1;
	int numrots;
	*/
	long longcolor;
	int /*len_cga, len_ega, len_vga,*/ len;
	byte /*xs,*/ xsb/*, ys*/;
	/*byte storetype;*/
	byte c;
	int x,y,b/*,flags*/;
	char *bucket,*tempbucket;
	unsigned char databyte,shapebyte;
	unsigned int /*colorand, colorshift,*/ colorcount;
	char *dest;
	int tblofs, dataofs;
	shapeheadertype sh;

	// K1n9_Duk3 mod: Read shape header from file, not from memory buffer.
	/*
	readin (numshapes, addr, 1);
	readin (numrots, addr, 2);
	readin (len_cga, addr, 2);
	readin (len_ega, addr, 2);
	readin (len_vga, addr, 2);
	readin (numcolorbits, addr, 1);
	readin (flags, addr, 2);
	*/
	read(handle, &sh, sizeof(sh));

	len=sh.len_vga+16;
	// K1n9_Duk3 mod: colorand and colorshift are constants here, so let's turn
	// them into macros to allow the compiler to optimize the code a bit:
#define colorand 255
#define colorshift 16
	/*
	colorand=255;
	colorshift=16;
	*/
	
	// K1n9_Duk3 mod: Using the numcolors macro would be wasteful, since it
	// would need to do the calculation for every iteration of the for-loops
	// below. And it wouldn't work anyway because the variable numcolorbits
	// is now part of a struct.
	colorcount = 1 << sh.numcolorbits;

	if (sh.flags & shm_fontf) {
		for (c=0; c<(colorcount); c++) colortab[c]=c;
		}
	else if (sh.numcolorbits==8) init8bit();
	else {
		for (c=0; c<colorcount; c++) {			// Numcolors is a macro
			read (handle, &longcolor, 4);
			colortab [c]=(longcolor>>colorshift)&colorand;
			};
		};
	
#undef colorand
#undef colorshift
	
	dest=malloc (len);
	if (dest==NULL) rexit (9);
	shm_tbllen[n]=len;
	shm_tbladdr[n]=dest;
	shm_flags[n]=sh.flags;
	tblofs=0;
	dataofs=sh.numshapes*4;
	
	// K1n9_Duk3 mod: Allocate temporary buffer in here, AFTER the buffer for
	// the final shape data was allocated, so we can avoid memory fragmentation.
	tempbucket = malloc(0x1000);	// use same size as the old bucket1 buffer
	if (!tempbucket) rexit(9);
	bucket = tempbucket;
  
	for (c=0; c<sh.numshapes; c++) {
		unsigned size;
		imageheadertype ih;
		// K1n9_Duk3 mod: Read image header from file, not from memory buffer.
		/*
		readin (xs, addr, 1);
		readin (ys, addr, 1);
		readin (storetype, addr, 1);
		*/
		read(handle, &ih, sizeof(ih));
		
		// K1n9_Duk3 mod: Let's make sure the image data actually fits into our
		// temporary buffer (to avoid memory corruption):
		size = ih.xs*ih.ys;
		if (size > 0x1000) rexit(42);
		
		// K1n9_Duk3 mod: Read image data from file instead of copying from
		// memory buffer. And since the bucket pointer won't change anyway,
		// there is no need to constantly re-assign it here.
		if (ih.storetype==st_byte) {
			read(handle, tempbucket, size);
			}
		/*
		bucket=bucket1;
		//	Copy the actual pixel bitmap of the shape into bucket
		if (storetype==st_byte) {
			memcpy (bucket, addr, xs*ys);
			(char*) addr+=(xs*ys);
			};
		*/

//	Now the shape definition is in memory at BUCKET
//	Translate to our mode:
		if ((sh.numcolorbits==8)&&(ih.xs==64)&&(ih.ys==12)) {
			memmove (&vgapal,bucket,sizeof (vgapal));
			vga_setpal();
			};

		xsb=ih.xs;
		wr (dest, tblofs, dataofs, 2);
		wr1 (dest, tblofs, xsb);
		wr1 (dest, tblofs, ih.ys);

		if (sh.flags & shm_blflag) {
			for (b=3; b>=0; b--) {
				for (y=0; y<ih.ys; y++) {
					for (x=0; x<ih.xs; x+=4) {
						shapebyte=colortab[*(byte*)(bucket+(x+b)+y*ih.xs)];
						wr1 (dest,dataofs,shapebyte);
						};
					};
				};
			}
		else {
			for (y=0; y<ih.ys; y++) {
				for (x=0; x<ih.xs; x++) {
					shapebyte=colortab[(*(byte*)(bucket+x+y*ih.xs))];
					wr1 (dest, dataofs, shapebyte);
					};
				};
			};
		};
	free(tempbucket);
	};

void shm_do (void) {
	long shoffset[128];
	/*char *shaddr[128];*/
	int shlen[128];					// Dual purpose
	int shafile;
	int c;
	/*char*bucket1;*/
 
	// K1n9_Duk3 mod: We don't actually need the bucket1 buffer for anything,
	// so let's get rid of it! This helps us avoid memory fragmentation.
	/*
	bucket1=malloc (4096);
	if (bucket1==NULL) rexit (9);
	
	for (c=0; c<shm_maxtbls; c++) {
		shaddr [c] = NULL;
		};
	*/
  
	shafile=_open (shm_fname,O_BINARY|O_RDONLY);
	// Read 128 long ints ... offsets of shape files in big file
	if (!read (shafile,shoffset,128*4)) rexit(9);

	// Read 128 words (short ints) ... lengths of each shape file
	read (shafile,shlen,128*2);
 
	for (c=0; c<shm_maxtbls; c++) {
		if (!shm_want[c]) {
	// This shape file exists, but not needed now.  Purge it from memory
			if (shm_tbladdr[c]!=NULL) {
				free (shm_tbladdr[c]);
				shm_tbladdr[c]=NULL;
				};
			}
		else if ((shm_tbladdr[c]==NULL)&&(shlen[c]!=0)) {
	// Shape file 'c' is needed, but not loaded, so load it.
			lseek (shafile, shoffset [c], SEEK_SET);	// Seek to position in file
			// K1n9_Duk3 mod: Don't allocate or read anything in here, just pass
			// the file handle to the xlate_table function and let it take care
			// of everything.
			/*
			shaddr[c]=malloc (shlen[c]);					// Allocate mem to hold it
			if (shaddr[c]==NULL) rexit (9);
			read (shafile, shaddr[c], shlen[c]);		// Read it in
			*/
			
			xlate_table (shafile,c);
			};
		};

	close (shafile);
	/*
	for (c=0; c<shm_maxtbls; c++) {
		if (shaddr[c]!=NULL) {
			xlate_table (c,shaddr[c], bucket1);
			free (shaddr[c]);
			};
		};
	free (bucket1);
	*/
	};

void shm_exit (void) {
	int c;
	for (c=0; c<shm_maxtbls; c++) {
		if (shm_tbladdr[c]!=NULL) {
			free (shm_tbladdr[c]);
			shm_tbladdr[c]=NULL;
			};
		};
	};